/* Tweaks (GTK 3) */

treeview {
    /* Set spacing for dropdown menu/entry completion items */
    -GtkTreeView-horizontal-separator: 12;
    -GtkTreeView-vertical-separator: 5;
}

filechooser treeview,
fontchooser treeview {
    /* Restore default item spacing in GTK choosers */
    -GtkTreeView-horizontal-separator: 2;
    -GtkTreeView-vertical-separator: 2;
}

.treeview-spacing {
    /* Disable GTK's built-in item spacing in custom treeviews */
    -GtkTreeView-horizontal-separator: 0;
    -GtkTreeView-vertical-separator: 0;
}

.dropdown-scrollbar {
    /* Enable dropdown list with a scrollbar */
    -GtkComboBox-appears-as-list: 1;
}

